#JES- Jython Environment for Students
#Copyright (C) 2002  Jason Ergle, Claire Bailey, David Raines, Joshua Sklare
#See JESCopyright.txt for full licensing information

import JESConstants
import java.awt as awt
import javax.swing as swing
import javax.swing.text.DefaultStyledDocument as DefaultStyledDocument
import keyword

WORD_BREAKS = [' ', '\n', '\t', '[', ']', '{', '}', ',', '\'', '-', '+', '=',
               '<', '>',  ':', ';', '_', '(', ')', '.',  '#', '"', '%' ]
KEYWORD_BOLD = 1
INSERT_EVENT = 1
REMOVE_EVENT = 2
MAX_UNDO_EVENTS_TO_RETAIN = 500

ERROR_LINE_FONT_COLOR       = awt.Color.white
ERROR_LINE_BACKGROUND_COLOR = awt.Color.red

class JESEditorDocument(DefaultStyledDocument):
################################################################################
# Function name: __init__
# Parameters:
#     -editor: JESEditor object that this object is associated with
# Return:
#     An instance of the JESEditorDocument class.
# Description:
#     Creates an instance of the JESEditorDocument class.
################################################################################
    def __init__(self, editor):
        self.editor = editor
        self.textAttrib = swing.text.SimpleAttributeSet()
        self.keywordAttrib = swing.text.SimpleAttributeSet()
        self.errorLineAttrib = swing.text.SimpleAttributeSet()
        swing.text.StyleConstants.setFontSize(self.textAttrib,
                                              JESConstants.FONT_SIZE)
        swing.text.StyleConstants.setForeground(self.keywordAttrib,
                                                JESConstants.KEYWORD_COLOR)
        swing.text.StyleConstants.setBold(self.keywordAttrib,
                                          KEYWORD_BOLD)
        swing.text.StyleConstants.setFontSize(self.keywordAttrib,
                                              JESConstants.FONT_SIZE)

        swing.text.StyleConstants.setForeground(self.errorLineAttrib,
                                                ERROR_LINE_FONT_COLOR)
        swing.text.StyleConstants.setBackground(self.errorLineAttrib,
                                                ERROR_LINE_BACKGROUND_COLOR)
        swing.text.StyleConstants.setFontSize(self.errorLineAttrib,
                                              JESConstants.FONT_SIZE)
        self.undoEvents = []

        #The following variables are set when showErrorLine is called.  They
        #are then used to unhighlight the line when the next text modification
        #is made.
        self.errorLineStart = 0
        self.errorLineLen   = 0

################################################################################
# Function name: insertString
# Parameters:
#     -offset: offset where the text will be inserted
#     -str: string that is being inserted
#     -a: attribute for the text that is being innserted.
# Description:
#     This function overrides the inherited insertString function.  It inserts
#     the target text and then calls the keywordHighlightEvent function to
#     highlight keywords.
################################################################################
    def insertString(self, offset, str, a, addUndoEvent=1):
        self.editor.modified = 1
        if addUndoEvent:
            self.addUndoEvent(INSERT_EVENT,
                              offset,
                              str)
        DefaultStyledDocument.insertString(self, offset, str, self.textAttrib)
        self.keywordHighlightEvent(offset, len(str))

################################################################################
# Function name: remove
# Parameters:
#     -offset: offset of the text that is being removed
#     -len: length of the text that is being removed
# Description:
#     This function overrides the inherited remove function.  It removes the
#     target text and then calls the keywordHighlightEvent function to highlight
#     keywords.
################################################################################
    def remove(self, offset, len, addUndoEvent=1):
        self.editor.modified = 1
        if addUndoEvent:
            self.addUndoEvent(REMOVE_EVENT,
                              offset,
                              self.getText(offset, len))
        DefaultStyledDocument.remove(self, offset, len)
        self.keywordHighlightEvent(offset, len)

################################################################################
# Function name: keywordHighlightEvent
# Parameters:
#     -modifiedTextOffset:
#     -modifiedTextLen:
# Description:
#     This function is called when text is either inserted or removed from the
#     document.  It takes in the text that was modified, and then calculates
#     where the first word before that text begins as well as where the next
#     word after that text ends.  This can then be used when calling the
#     updateKeywordHighlightInRange function to let it know where to update
#     keyword highlights.
################################################################################
    def keywordHighlightEvent(self, modifiedTextOffset, modifiedTextLen):
        #Unhighlight a line if showErrorLine was called earlier
        if self.errorLineStart > 0:
            self.updateKeywordHighlightInRange(self.errorLineStart,
                                               self.errorLineLen)
            self.errorLineStart = 0
            self.errorLineLen   = 0

        #Find start of the word before the modified text
        startOffset = modifiedTextOffset - 1
        while startOffset > 0:
            startOffset -= startOffset
            curChar = self.getText(startOffset, 1)
            if curChar in WORD_BREAKS:
                break

        #Find end of the word after the modified text
        endOffset = modifiedTextOffset + modifiedTextLen
        while endOffset <= self.length:
            curChar = self.getText(endOffset, 1)
            if curChar in WORD_BREAKS:
                break
            endOffset += 1

        #Ensure that the start and end offsets are within vaild document range
        if endOffset >= self.length:
            endOffset = self.length
        if startOffset < 0:
            startOffset = 0

        self.updateKeywordHighlightInRange(startOffset, endOffset - startOffset)

################################################################################
# Function name: updateKeywordHighlightInRange
# Parameters:
#     -offset: offset where text begins that needs to be checked for keywords
#     -len: length of the text which needs to be checked for keywords
# Description:
#     This function takes in a text offset and length, searches for keywords in
#     the text, and highlights those keywords.
#     It does this by first removing the entire text and replacing it with
#     default attribute text.  This ensures that any text that once was a
#     keyword will not be highlighted anymore.  Then the text is searched for
#     keywords;  when a keyword is found, that text is replaced with the keyword
#     highlighted text.
################################################################################
    def updateKeywordHighlightInRange(self, offset, len):
        #Save the cursor position so it can be reset later
        curOffset = self.editor.getCaretPosition()
        rangeText = self.getText(offset, len)

        #Delete all text within range and replace with default style text
        DefaultStyledDocument.remove(self, offset, len)
        DefaultStyledDocument.insertString(self,
                                           offset,
                                           rangeText,
                                           self.textAttrib)
        i = 0
        iWordStart = 0
        curText = ''

        #Read each character in the range until an entire word is found
        while i < len:
            curChar = rangeText[i]

            #If the currect char is a word break char, then an entire word has
            #been read
            if curChar in WORD_BREAKS:
                self.setTextAttrib(offset + iWordStart, curText)
                iWordStart = i + 1
                curText = ''
            else:
                curText += curChar
            i += 1

        self.setTextAttrib(offset + iWordStart, curText)
        self.editor.setCaretPosition(curOffset)

################################################################################
# Function name: setTextAttrib
# Parameters:
#     -offset: offset where text begins the will be updated
#     -text: text that will be updated
# Description:
#     This function is called set the text attribute of the specified text.  It
#     will check to see if that text is a keyword, and if so, will set the
#     text attribute so that the word has the correct keyword highlighting.
################################################################################
    def setTextAttrib(self, offset, text):
        if keyword.iskeyword(text):
            DefaultStyledDocument.remove(self, offset, len(text))
            DefaultStyledDocument.insertString(self,
                                               offset,
                                               text,
                                               self.keywordAttrib)

################################################################################
# Function name: addUndoEvent
# Parameters:
#     -eventType: identifies the type of event that occured (insert or remove)
#     -offset: offset in the text that the event occured in
#     -str: text that is being inserted or removed
# Description:
#     Adds an undo event to the event array.  This is needed so that text
#     modification events can be undone.  If the array reaches it's maximum
#     capacity, then the oldest event is removed from the array before adding
#     the new one.
################################################################################
    def addUndoEvent(self, eventType, offset, str):
        if len(self.undoEvents) > MAX_UNDO_EVENTS_TO_RETAIN:
            del self.undoEvents[0]
            
        self.undoEvents.append([eventType,
                                offset,
                                str])

################################################################################
# Function name: undo
# Description:
#     Undoes the last text modification that is in the undo events array and
#     removes it from the array.
################################################################################
    def undo(self):
        if len(self.undoEvents) > 0:
            lastEvent = self.undoEvents.pop()
            if lastEvent[0] == INSERT_EVENT:
                self.remove(lastEvent[1],
                            len(lastEvent[2]),
                            0)
            else:
                self.insertString(lastEvent[1],
                                  lastEvent[2],
                                  self.textAttrib,
                                  0)

################################################################################
# Function name: showErrorLine
# Parameters:
#      -lineNumber: number of the line that should be highlighted
# Description:
#     When this function is called, the specified line will be highlighted so
#     that the user can tell which line contains an error.
################################################################################
    def showErrorLine(self, lineNumber):
        #Search for the start offset of the error line
        docText = self.getText(0, self.getLength())
        line   = 1
        offset = 0
        while line < lineNumber:
            offset = docText.find('\n', offset)
            line += 1

        #Search for the end offset of the error line
        offset += 1
        endOffset = docText.find('\n', offset)

        #Set error line position and length object variables
        self.errorLineStart = offset
        self.errorLineLen   = endOffset - offset

        #Set the correct text attribute for the error line
        self.setCharacterAttributes(self.errorLineStart,
                                    self.errorLineLen,
                                    self.errorLineAttrib,
                                    0)

        #Set cusor to error line to ensure that the error line will be visible
        self.editor.setCaretPosition(self.errorLineStart)
